> module InXmlTest where

> -- import View
> import Inv
> import InvPrelude
> import Val hiding (PNPair(..))
> import Eval   -- for testing only
> import InXml
> import Data.Tree


some simple tests for filters implemented by Inv are presented.
there is an a little big example filter in the middle of this file.

t1, t2, t3 and t4 are test inputs used below.

> t1 = "{'a', '1':'2':'3':[]}"
> t2 = "'a'"
> t3 = "{'a', {'b', '1':'2':'3':[]}:{'c', '4':'5':'6':[]}:{'b', '7':'8':'9':[]}:[]}"
> t4 = "{'a', {'b', '1':'2':'3':[]}:{'a', {'a', '10':[]}:'5':'6':[]}:{'b', {'a', '11':[]}:'8':'9':[]}:[]}"

Predicates
----------

none

> nonetest = ev (none) t1


keep

> keeptest = ev (keep) t1


elm

> elmtest1 = ev (elm) t1
> elmtest2 = ev (elm) t2


txt

> txttest1 = ev (txt) t1
> txttest2 = ev (txt) t2

tag

> tagtest1 = ev (tag "a") t1
> tagtest2 = ev (tag "b") t1
> tagtest3 = ev (tag "a") t2



Selection
---------

children

> childrentest1 = ev (children) t1
> childrentest2 = ev (children) t2


Construction
------------

literal

> literaltest = ev (literal "text") t1

replaceTag

> replaceTagtest1 = ev (replaceTag "tag") t1
> replaceTagtest2 = ev (replaceTag "tag") t2


mkElem

> mkElemtest1 = ev (mkElem "ele" [children, keep]) t1
> mkElemtest2 = ev (mkElem "ele" [children, keep]) t2


Combinators
-----------

f `o` g

> icomptest = ev (children `o` (mkElem "ele" [children, keep])) t1


f ||| g

> uniontest = ev (children ||| (mkElem "ele" [children, keep])) t1


cat

> cattest = ev (cat [children,(mkElem "ele" [children, keep])]) t1


f `with` g

> withtest = (ev (children `with` tag "b")) t3

f `without` g

> withouttest = (ev (children `without` tag "b")) t3


f /> g

> indirtest1 = (ev (tag "a" /> tag "b")) t3
> indirtest2 = (ev (tag "a" /> tag "a")) t3


f </ g

> exdirtest1 = (ev (tag "a" </ tag "b")) t3
> exdirtest2 = (ev (tag "a" </ tag "a")) t3


p ?> f :> g

> iftest1 = (ev (tag "a" ?> children :> none)) t3
> iftest2 = (ev (tag "b" ?> children :> none)) t3


f |>| g

> distest1 = (ev (txt |>| children)) t1
> distest2 = (ev (txt |>| children)) t2


chip f

> chiptest1 = (ev (chip (replaceTag "x"))) t3



deep f

> deeptest1 = (ev (deep (tag "a"))) t4


deepest

> deepesttest1 = (ev (deepest (tag "a"))) t4


multi f

> multitest1 = (ev (multi children)) t4


foldXml f

> foldtest1 = (ev (foldXml (replaceTag "x" |>| keep))) t3



------------------- middle size example

> html  = mkElem "html"
> body  = mkElem "body"
> h1    = mkElem "h1"
> ul    = mkElem "ul"
> li    = mkElem "li"
> table = mkElem "table"
> tr    = mkElem "tr"
> th    = mkElem "th"
> td    = mkElem "td"

an example filter 
now obselete

> addrtable = 
>  html 
>   [ body 
>     [ h1 [ literal "IPLAddressBook" ],
>       ul [replaceTag "li" `o` (keep /> tag "person" /> tag "name") ],
>		table 
>		  [tr [ th [literal "Name"], th [literal "Email"], th [literal "Tel"]],
>	       foldXml tabling `o` (keep /> tag "person")]]]
> tabling = tag "person" ?> replaceTag "tr" :>
>		    (tag "name" ?> replaceTag "td" :>
>		    (tag "email" ?> replaceTag "td" :>
>		    (tag "tel" ?> replaceTag "td" :>
>			keep)))


original data(general tree which is used in X)

> addrBook = 
>     Data.Tree.Node "addrbook" 
>      [Data.Tree.Node "person" p1, Data.Tree.Node "person" p2, Data.Tree.Node "person" p3, Data.Tree.Node "person" p4]
>  where
>  p1 = [Data.Tree.Node "name" [Data.Tree.Node "Zhenjiang Hu" []],
>	     Data.Tree.Node "email" [Data.Tree.Node "hu@mist.i.u-tokyo.ac.jp" []],
>	     Data.Tree.Node "tel" [Data.Tree.Node "+81-3-5841-7411" []]]
>  p2 = [Data.Tree.Node "name" [Data.Tree.Node "Kento Emoto" []],
>	     Data.Tree.Node "email" [Data.Tree.Node "emoto@ipl.t.u-tokyo.ac.jp" []],
>	     Data.Tree.Node "tel" [Data.Tree.Node "+81-3-5841-7412" []]]
>  p3 = [Data.Tree.Node "name" [Data.Tree.Node "Shin-Cheng Mu" []],
>	     Data.Tree.Node "email" [Data.Tree.Node "scm@mist.i.u-tokyo.ac.jp" []],
>	     Data.Tree.Node "tel" [Data.Tree.Node "+81-3-5841-7411" []]]
>  p4 = [Data.Tree.Node "name" [Data.Tree.Node "Masato Takeichi" []],
>	     Data.Tree.Node "email" [Data.Tree.Node "takeichi@acm.org" []],
>	     Data.Tree.Node "tel" [Data.Tree.Node "+81-3-5841-7430" []]]



a function converting a tree in X into string expressing a tree in Inv

> convxi x = convxi' x []
>  where
>  convxi' (Data.Tree.Node x []) = ((ss x)++)
>  convxi' (Data.Tree.Node x xs) = ("{"++).((ss x)++).(","++).(foldl1 (cc) (map convxi' xs)).(":[]}"++)
>  cc x y = x.(":"++).y
>  ss x = "'" ++ init(tail(show x)) ++"'"


addrBook converted into Inv's expression

> addrBook2 = "{'addrbook',{'person',{'name','Zhenjiang Hu':[]}:{'email','hu@mist.i.u-tokyo.ac.jp':[]}:{'tel','+81-3-5841-7411':[]}:[]}:{'person',{'name','Kento Emoto':[]}:{'email','emoto@ipl.t.u-tokyo.ac.jp':[]}:{'tel','+81-3-5841-7412':[]}:[]}:{'person',{'name','Shin-Cheng Mu':[]}:{'email','scm@mist.i.u-tokyo.ac.jp':[]}:{'tel','+81-3-5841-7411':[]}:[]}:{'person',{'name','Masato Takeichi':[]}:{'email','takeichi@acm.org':[]}:{'tel','+81-3-5841-7430':[]}:[]}:[]}"

> vfst (a :& b) = a
> vsnd (a :& b) = b
> vmap f Nl = []
> vmap f (a :@ b) = (f a):(vmap f b)


a function converting a tree in Inv into a tree in X

> convix x = convix' x
>  where
>  convix' (Str s) = (Data.Tree.Node s [])
>  convix' (Nod a x) = (Data.Tree.Node (convixs' a) (convixl' x))
>  convixl' Nl = []
>  convixl' (a :@ x) = ((convix' a):(convixl' x))
>  convixs' (Str s) = s


ipladdrtable is an HTML obtained by applying the example transformation addrtable to addrBook.

> ipladdrtable = do
>  a <- ev (addrtable) (convxi addrBook)
>  return (trans2XML 0 (head (vmap convix (vsnd a))) "")

> getCont (Right a) = a

> prt = printlines (linize (getCont ipladdrtable))


auxiliary functions to display HTML

> trans2XML k (Data.Tree.Node n []) = (tabify k).(n ++).("\n"++)
> trans2XML k (Data.Tree.Node n xs) = (tabify k).("<"++).(n++).(">\n"++).
>			     			(foldr (.) id (map (trans2XML (k+1)) xs)).
>			                (tabify k).("</"++).(n++).(">\n"++)

> tabify k = ((foldr (++) "" (map (\x -> "  ") [1..k])) ++)

> getFirstChild (Data.Tree.Node n xs) = head xs


> linize :: String -> [String]
> linize = lin [] []
>  where 
>  lin xs cur [] = xs++[reverse cur]
>  lin xs cur ('\n':rest) = lin (xs++[reverse cur]) [] rest
>  lin xs cur (c:rest) = lin xs (c:cur) rest

> printlines = (foldr (>>) (putStr "")).(map putStrLn)

you can print HTML(XML) by executing printXML ipladdrtable

> printXML = printlines.linize


 ----

> fn = mkElem "m" [tag "a" `o'` children, children]
> t5 = "{'r',{'b',[]}:{'a',[]}:[]}"

*InXmlTest> ev fn t5
Right <{'r',{'b',[]}:{'a',[]}:[]},{'m',{'a',[]}:{'b',[]}:{'a',[]}:[]}:[]>

*InXmlTest> ev (Inv fn) "<{'r',{'b',[]}:{'a',[]}:[]},{'m',{'a',[]}:{'b',[]}:{'a',[]}:[]}:[]>"
Right {'r',{'b',[]}:{'a',[]}:[]}

*InXmlTest> ev (Inv fn) "<{'r',{'b',[]}:{'a',[]}:[]},{'m',{'a',[]}:\\({'b',[]}:{'a',[
]}:[])}:[]>"
Right {'r',\({'b',[]}:{'a',[]}:[])}


The Address Book Example
------------------------

> addrbook2' = "{'a',{'p',{'n','ZH':[]}:{'e','hu@mist':[]}:[]}:{'p',{'n','KE':[]}:{'e','emoto@ipl':[]}:[]}:{'p',{'n','SCM':[]}:{'e','scm@mist':[]}:[]}:[]}"

> mkView =
>   Define "mkView" 
>     (html [mkIdx, mkTable])

> mkIdx =
>   Define "mkIdx" (replaceTag "ul" `o'` chip' mkLi)
>  where mkLi = -- Define "mkLi" (replaceTag "li" `o'` (tag "p" //> tag "n"))
>          Prim (\st -> eval st (replaceTag "li" `o'`(tag "p" //> tag "n")),
>                mkl)
>        mkl _ (_ :& (Nod _ v:@ Nl))
>          = return (Nod (Str "p") (Nod (Str "n") v :@ Undef :@ Nl))
>        mkl _ (Ins (_ :& (Nod _ v:@ Nl)))
>          = return (Ins (Nod (Str "p") (Nod (Str "n") v :@ Undef :@ Nl)))
>        mkl _ (Del (_ :& (Nod _ v:@ Nl)))
>          = return (Del (Nod (Str "p") (Nod (Str "n") v :@ Undef :@ Nl)))
>        mkl st v = eval st (Inv (replaceTag "li" `o'`(tag "p" //> tag "n"))) v

> mkTable =
>   Define "mkTable" (replaceTag "table" `o'` chip' mkRow)

> mkRow =
>   Define "mkRow"
>     (mkElem' "tr" [f1, f2])

> f1 = Prim (\st -> eval st (replaceTag "td" `o'`(tag "p" //> tag "n")),
>            f1')
>  where f1' _ (Nod (Str s) Undef :& (Nod _ v:@ Nl))
>          | s == "p" = return (Nod (Str "p") (Nod (Str "n") v :@ Undef :@ Nl))
>        f1' st v = eval st (Inv (replaceTag "td" `o'`(tag "p" //> tag "n"))) v
> f2 = Prim (\st -> eval st (replaceTag "td" `o'`(tag "p" //> tag "e")),
>            f2')
>  where f2' _ (Nod (Str s) Undef :& (Nod _ v:@ Nl))
>          | s == "p" = return (Nod (Str "p") (Undef :@ Nod (Str "e") v :@ Nl))
>        f2' st v = eval st (Inv (replaceTag "td" `o'`(tag "p" //> tag "e"))) v

> f //> g = g `o'` children `o` f


> mkElem' s fs = 
>  Define ("mkElem "++s++" ["++ concat (map show fs) ++"]")
>    (applyfilters fs <.> (Id :*: concatx'') 
>          <.> Assocl <.> (Inv (applyfilters fs) :*: Id) 
>          <.> (Id :*: Dup (DStr s) <.> Swap <.> Inv.Node <.> wrap))
>  where concatx'' = Prim (\st -> eval st concatx, 
>                          \_ (_ :& x) -> return (singletons x))
>        singletons Nl = Nl
>        singletons (a :@ x) = (a :@ Nl) :@ singletons x


TODO: Resolve:

value <{'p',_|_},{'td','KE':[]}:[]> rejected by f1^o because
value <[],{'td','KE':[]}:[]> rejected by concatx'^o because
value [] not in the domain of cons^o